# -*- coding: utf-8 -*-

"""Top-level package for openwrt-luci-rpc."""

__author__ = """Finbarr Brady"""
__email__ = 'fbradyirl@github.io'
__version__ = '1.1.7'

from .openwrt_luci_rpc import OpenWrtLuciRPC
from .constants import Constants


class OpenWrtRpc:
    """
    Class to interact with OpenWrt router running luci-mod-rpc package.
    """

    def __init__(self, host_url=Constants.DEFAULT_LOCAL_HOST,
                 username=Constants.DEFAULT_USERNAME,
                 password=Constants.DEFAULT_PASSWORD,
                 is_https=Constants.DEFAULT_HTTPS,
                 verify_https=Constants.DEFAULT_VERIFY_HTTPS):
        """
        Initiate an instance with a default local ip (192.168.1.1)
        :param host_url: string - host url. Defaults to 192.168.1.1
        :param username: string - username. Defaults to root
        :param password: string - password. Default is blank
        :param is_https: boolean - use https? Default is false
        :param verify_https: boolean - verify https? Default is true
        """
        self.router = OpenWrtLuciRPC(host_url, username, password,
                                     is_https, verify_https)

    def is_logged_in(self):
        """Returns true if a token has been aquired"""
        return self.router.token is not None

    def get_all_connected_devices(self,
                                  only_reachable=Constants.DEFAULT_ONLY_REACH,
                                  wlan_interfaces=Constants.DEFAULT_WLAN_IF):
        """Get details of all devices"""
        return self.router.get_all_connected_devices(
            only_reachable=only_reachable, wlan_interfaces=wlan_interfaces)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `openwrt_luci_rpc` package."""


import unittest
from openwrt_luci_rpc import utilities
from openwrt_luci_rpc.constants import Constants


class TestOpenwrtLuciRPC(unittest.TestCase):
    """Tests for `openwrt_luci_rpc` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_normalise_keys(self):
        """Test replacing v18 keys works as expected."""

        data = {'dest': "10.1.1.11"}

        data = utilities.normalise_keys(data)

        assert data[Constants.MODERN_KEYS["dest"]] == '10.1.1.11'



class Constants:

    DEFAULT_LOCAL_HOST = "192.168.1.1"
    DEFAULT_USERNAME = "root"
    DEFAULT_PASSWORD = ""
    DEFAULT_TIMEOUT = 30
    DEFAULT_HTTPS = False
    DEFAULT_VERIFY_HTTPS = True
    DEFAULT_ONLY_REACH = True
    DEFAULT_WLAN_IF = ['wlan0-1']

    LUCI_RPC_LOGIN_PATH = '{}/cgi-bin/luci/rpc/auth'
    LUCI_RPC_SYS_PATH = '{}/cgi-bin/luci/rpc/sys'
    LUCI_RPC_UCI_PATH = '{}/cgi-bin/luci/rpc/uci'

    # Only available in 18.06+
    LUCI_RPC_IP_PATH = '{}/cgi-bin/luci/rpc/ip'

    # Make old and new keys consistent for 17 & 18 releases
    # Spaces and dots are removed so keep them out of here also.
    MODERN_KEYS = {
        "HW_address": "mac",
        "IP_address": "ip",
        "dest": "ip"
    }

    def __init__(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `openwrt_luci_rpc` package."""


import unittest
from unittest.mock import Mock, patch

from openwrt_luci_rpc import OpenWrtRpc
from openwrt_luci_rpc.constants import Constants
from openwrt_luci_rpc import utilities
from openwrt_luci_rpc.exceptions import LuciConfigError


class TestOpenwrtLuciRPC(unittest.TestCase):
    """Tests for `openwrt_luci_rpc` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_none_host(self):
        """Test invalid host raises exception."""
        with self.assertRaises(LuciConfigError):
            OpenWrtRpc(None)

    def test_empty_host(self):
        """Test invalid host raises exception."""
        with self.assertRaises(LuciConfigError):
            OpenWrtRpc("")

    @patch('requests.Session.post')
    def test_defaults(self, mock_post):
        """Test defaults are in place."""

        json_result = {
            "id": "1",
            "result": "2e9f2c44d34a10e1321de7893123123221",
            "error": None
        }

        mock_post.return_value = Mock(ok=True)
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = json_result

        runner = OpenWrtRpc()
        assert runner.router.host_api_url == '{}://{}'.format(
                    "http",
                    Constants.DEFAULT_LOCAL_HOST
                )
        assert runner.router.username == Constants.DEFAULT_USERNAME
        assert runner.router.password == Constants.DEFAULT_PASSWORD

    def test_normalise_key_stripping(self):
        """Test replacing dots and spaces works."""

        data = {
            ".name": "cfg07ee1",
            ".type": "host",
            "name": "imac-ethernet",
            ".index": 4,
            "mac": "c8:2a:10:4a:10:dd",
            "dns": "1",
            ".anonymous": True,
            "ip": "192.168.1.124"
        }

        data = utilities.normalise_keys(data)

        assert data['_name'] == "cfg07ee1"
        assert data['_type'] == "host"
        assert data['ip'] == "192.168.1.124"

    def test_normalise_key_none_type(self):
        """Test a None result is handled."""

        data = utilities.normalise_keys(None)

        assert data is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `openwrt_luci_rpc` package."""

import unittest
import os
from openwrt_luci_rpc.openwrt_luci_rpc import OpenWrtLuciRPC


class TestOpenwrt15LuciRPC(unittest.TestCase):

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    @unittest.skip("unskip this to test using env vars")
    def testDiscover(self):
        # HOST, USER, PASSWORD,[HTTPS] must be in env variables in order
        # to run the test
        assert "HOST" in os.environ
        assert "USER" in os.environ
        assert "PASSWORD" in os.environ

        router = OpenWrtLuciRPC(
                    os.getenv("HOST"),
                    os.getenv("USER"),
                    os.getenv("PASSWORD"),
                    os.getenv("HTTPS", "False") == "True",
                    os.getenv("VERIFY_HTTPS", "False") == "True",
                )
        devices = router.get_all_connected_devices(False, False)
        assert devices is not None


if __name__ == '__main__':
    unittest.main()



from .constants import Constants
import logging

log = logging.getLogger(__name__)


def normalise_keys(result):
    """Replace 17.06 keys with newer ones."""
    if result is not None:

        # named tuple keys cannot have spaces or begin with dots. Replace
        # those with underscores
        result = \
            {k.replace(".", "_").replace(" ", "_"): v for k, v in result.items()}  # noqa: E501

        for old_key, new_key in Constants.MODERN_KEYS.items():
            if old_key in result:
                result[new_key] = result[old_key]
                del result[old_key]

    return result


def get_hostname_from_dhcp(dhcp_result, mac):
    """Determine the hostname for this mac."""
    if dhcp_result is not None and isinstance(dhcp_result, dict):
        host = [x for x in dhcp_result.values()
                if x['.type'] == 'host'
                and 'mac' in x
                and 'name' in x
                and x['mac'].upper() == mac]

        if host:
            log.debug("DNS name lookup for mac {} "
                      "found {}".format(mac, host[0]['name']))
            return host[0]['name']

    return None



# -*- coding: utf-8 -*-

"""
Support for OpenWrt (luci) routers.

For more details about this platform, please refer to the documentation at
https://home-assistant.io/components/device_tracker.luci/
"""

import requests
import json
import logging

from packaging import version
from collections import namedtuple
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from openwrt_luci_rpc import utilities
from .constants import Constants
from .exceptions import InvalidLuciTokenError, \
    LuciRpcMethodNotFoundError, InvalidLuciLoginError, \
    LuciRpcUnknownError, PageNotFoundError, LuciConfigError

log = logging.getLogger(__name__)


class OpenWrtLuciRPC:

    def __init__(self, host, username, password, is_https, verify_https):
        """
        Initiate an API request with all parameters
        :param host: string
        :param username: string
        :param password: string
        :param is_https: boolean
        :param verify_https: boolean
        """

        if not host:
            raise LuciConfigError('host cannot be empty. '
                                  'Use the IP or hostname of your'
                                  'OpenWrt router')

        if not is_https:
            protocol = 'http'
        else:
            protocol = 'https'

            if not verify_https:
                requests.packages.urllib3.disable_warnings(
                    InsecureRequestWarning
                )

        self.host = host
        self.host_api_url = '{}://{}'.format(protocol, host)
        self.username = username
        self.password = password
        self.verify_https = verify_https
        self.session = requests.Session()
        self.token = None
        self.owrt_version = None
        self._refresh_token()
        self.is_legacy_version, self.arp_call \
            = self._determine_if_legacy_version()

    def _refresh_token(self):
        """Get authentication token for the given configuration."""
        auth_url = Constants.\
            LUCI_RPC_LOGIN_PATH.format(self.host_api_url)
        self.token = self._call_json_rpc(auth_url, 'login',
                                         self.username, self.password)
        log.info("Luci RPC login was successful")

    def _determine_if_legacy_version(self):
        """
        * Checks to see if we are running a pre-18.06 version

        :return: tuple, with True if legacy and the URL to
                use to lookup devices
        """

        log.info("Getting OpenWRT version")

        # NEW METHOD TO DETERMINE OPENWRT VERSION EXACTLY
        # TODO: Check as non-root user

        # get VERSION_ID from os-release if exists or get
        # DISTRIB_RELEASE from openwrt_release
        shell_command = "if [ -f \"/etc/os-release\" ]; \
                            then awk -F= '$1==\"VERSION_ID\" \
                            { print $2 ;}' \
                            /etc/os-release; \
                            else awk -F= '$1==\"DISTRIB_RELEASE\" \
                            { print $2 ;}' \
                            /etc/openwrt_release; fi | \
                            tr -d \\'\\\""

        rcp_sys_version_call = Constants.\
            LUCI_RPC_SYS_PATH.format(self.host_api_url), "exec"

        try:
            content = self._call_json_rpc(rcp_sys_version_call[0],
                                          rcp_sys_version_call[1],
                                          shell_command)   # type: str

            content = content.replace("\n", "")

            if content is None:
                raise LuciRpcUnknownError("could not \
                determine openwrt version")

            self.owrt_version = version.parse(content.strip())
        except InvalidLuciLoginError:
            log.info("Refreshing login token")
            self._refresh_token()
            return self._determine_if_legacy_version()
        except Exception:
            log.error("Could not determine OpenWRT version, \
                         defaulting to version 18.06")
            self.owrt_version = version.parse("18.06")

        rpc_sys_arp_call = Constants.\
            LUCI_RPC_SYS_PATH.format(self.host_api_url), 'net.arptable'
        rpc_ip_call = Constants.\
            LUCI_RPC_IP_PATH.format(
                self.host_api_url), 'neighbors', {"family": 4}

        if self.owrt_version != version.parse("snapshot") and self.owrt_version < version.parse("18.06"):  # noqa: E501
            return True, rpc_sys_arp_call
        else:
            return False, rpc_ip_call

    def get_all_connected_devices(self, only_reachable, wlan_interfaces):
        """
        Get details of all connected devices.

        Notes around newer OpenWrt releases (18.06+)

            Do not use `reachable` or `stale` values
            as they flap constantly even
            when the device is inside the network.
            The very existence of the mac in the results
            is enough to determine the "device is home"

        :param only_reachable: boolean, if true,
               only return devices which are reachable
               (this is for 17.06 or earlier only. 18+
               does not have a proper ability to determine
               this, as above)
        """
        log.debug("Checking for connected devices")
        last_results = []
        # rpc_sys__winfo_call = Constants.\
        #     LUCI_RPC_SYS_PATH.format(self.host_api_url), \
        #                       'wifi.getiwinfo', wlan_interfaces
        rpc_uci_call = Constants.LUCI_RPC_UCI_PATH.format(
            self.host_api_url), 'get_all', 'dhcp'

        try:
            # First, try find the associated wifi devices
            # winfo_result = self._call_json_rpc(*rpc_sys__winfo_call)
            arp_result = self._call_json_rpc(*self.arp_call)
            dhcp_result = self._call_json_rpc(*rpc_uci_call)
        except InvalidLuciTokenError:
            log.info("Refreshing login token")
            self._refresh_token()
            return self.get_all_connected_devices(only_reachable,
                                                  wlan_interfaces)

        for device_entry in arp_result:
            if device_entry is None:
                continue

            device_entry = utilities.normalise_keys(device_entry)

            if "mac" not in device_entry:
                continue

            device_entry['hostname'] = utilities.get_hostname_from_dhcp(
                dhcp_result, device_entry['mac'])

            # As a convenience, add the router IP as the host
            # for every device. Can be useful when a network has more
            # than one router.
            if "host" not in device_entry:
                device_entry['host'] = self.host

            device = namedtuple("Device", device_entry.keys())(
                *device_entry.values())

            if "Flags" in device_entry and only_reachable:
                # Check if the Flags for each device contain
                # NUD_REACHABLE and if not, skip.
                if not int(device_entry['Flags'], 16) & 0x2:
                    continue

            last_results.append(device)

        log.debug(last_results)
        return last_results

    def _call_json_rpc(self, url, method, *args, **kwargs):
        """Perform one JSON RPC operation."""
        data = json.dumps({'method': method, 'params': args})

        # pass token to make it work with versions < 17
        if self.token is not None:
            url += "?auth=" + self.token

        log.debug("_call_json_rpc : %s" % url)
        res = self.session.post(url,
                                data=data,
                                timeout=Constants.DEFAULT_TIMEOUT,
                                verify=self.verify_https,
                                **kwargs)

        if res.status_code == 200:
            result = res.json()
            try:
                content = result['result']

                if content is not None:
                    return content

                elif result['error'] is not None:
                    # On 18.06, we want to check for error 'Method not Found'
                    error_message = result['error']['message']
                    error_code = result['error']['code']
                    if error_code == -32601:
                        raise LuciRpcMethodNotFoundError(
                            "method: '%s' returned an "
                            "error '%s' (code: '%s).",
                            method, error_message, error_code)
                else:
                    log.debug("method: '%s' returned : %s" % (method, result))
                    # Authentication error
                    raise InvalidLuciLoginError("Failed to authenticate "
                                                "with Luci RPC, check your "
                                                "username and password.")

            except KeyError:
                raise LuciRpcUnknownError("No result in response from luci")

        elif res.status_code == 401:
            raise InvalidLuciLoginError("Failed to authenticate "
                                        "with Luci RPC, check your "
                                        "username and password.")
        elif res.status_code == 403:
            raise InvalidLuciTokenError("Luci responded "
                                        "with a 403 Invalid token")
        elif res.status_code == 404:
            raise PageNotFoundError("404 returned "
                                    "from %s. Ensure you have "
                                    "installed package "
                                    "`luci-mod-rpc`." % url)

        raise LuciRpcUnknownError("Invalid response from luci: %s", res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `openwrt_luci_rpc` package."""


import unittest
from openwrt_luci_rpc import utilities
from openwrt_luci_rpc.constants import Constants


class TestOpenwrtLuciRPC(unittest.TestCase):
    """Tests for `openwrt_luci_rpc` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_normalise_keys(self):
        """Test replacing v17 keys works as expected."""

        data = {
                "HW type": "0x1",
                "Mask": "*",
                "Flags": "0x2",
                "Device": "br-lan",
                'HW address': '9C:20:7B:CA:A2:16',
                'IP address': "127.0.0.1",
                }

        data = utilities.normalise_keys(data)

        assert data[Constants.MODERN_KEYS["HW_address"]] == '9C:20:7B:CA:A2:16'
        assert data[Constants.MODERN_KEYS["IP_address"]] == '127.0.0.1'
        assert data['HW_type'] == "0x1"
        assert data['Mask'] == "*"
        assert data['Flags'] == "0x2"
        assert data['Device'] == "br-lan"
